from typing import Dict, List, Tuple, Set, Union

from tqdm import tqdm

class LabelTokenizer():
    
    def __init__(self, filename: str=None) -> None:
        self.idx2label: List[str] = []
        self.label2idx: Dict[str, int] = {}
        if filename is not None:
            with open(filename, 'r') as f:
                for idx, line in tqdm(enumerate(f), desc=f'Loading labels from {filename}'):
                    self.idx2label.append(line.strip())
                    self.label2idx[line.strip()] = idx  
                    self.__make_o_first()
                    
    def __make_o_first(self):
        if "O" in self.label2idx:
            # find O idx
            idx = self.label2idx["O"]
            # swap with first label
            self.label2idx[self.idx2label[0]], self.label2idx["O"] = self.label2idx["O"], self.label2idx[self.idx2label[0]]
            self.idx2label[0],self.idx2label[idx] = self.idx2label[idx], self.idx2label[0]
                    
    def load(self, labels: List[str], sort=True):
        if sort:
            labels = list(labels)
            labels.sort()
        for label in labels:
            self.idx2label.append(label)
        for idx, label in enumerate(self.idx2label):
            self.label2idx[label] = idx
        self.__make_o_first()
        return self
    
    def remove(self, index: int):
        label = self.idx2label[index]
        self.idx2label.remove(label)
        self.label2idx = {}
        for idx, label in enumerate(self.idx2label):
            self.label2idx[label] = idx
        return self
    
    def add(self, item, index=-1):
        if index!=-1:
            self.idx2label.insert(index, item)
        else:
            self.idx2label.append(item)
        for idx, label in enumerate(self.idx2label):
            self.label2idx[label] = idx
        return self
    
    def convert_tokens_to_ids(self, label: Union[str, List[str]]) -> int:
        if isinstance(label, list):
            return [self.label2idx[l] for l in label]
        return self.label2idx[label]
    
    def convert_ids_to_tokens(self, idx: Union[int, List[int]]) -> int:
        if isinstance(idx, list):
            return [self.idx2label[i] for i in idx]
        return self.idx2label[idx]
    
    def __len__(self):
        return len(self.idx2label)
    
class HashTrie():
    
    def __init__(self) -> None:
        self.trie = set()
        self.max_length = 0
        
    def load_vocab_from_file(self, filename: str):
        with open(filename, 'r') as f:
            for line in tqdm(f, desc='Loading vocab'):
                self.trie.add(line.strip())
                self.max_length = max(self.max_length, len(line.strip()))
                
    def __add__(self, word):
        self.trie.add(word)
        self.max_length = max(self.max_length, len(word))
        return self
                
    def search_words(self, sentence: str, words_number:int=5, return_list=True):
        if return_list:
            words = []
        else:
            words = set()
        for start in range(len(sentence)):
            if return_list:
                words.append([])
            s = ""
            for end in range(start, min(start+self.max_length, len(sentence))):
                s += sentence[end]
                if s in self.trie:
                    if return_list:
                        words[-1].append(s)
                        if len(words[-1])>=words_number:
                            break
                    else:
                        words.add(s)
        return words
    
    def __len__(self):
        return len(self.trie)

def is_span_intersect(a: Tuple[int,int], b: Tuple[int,int]):
    """
    Determine if two spans intersect.
    a[0]<=b[1] and b[0]<=a[1]

    Args:
        a (Tuple[int,int]): First span
        b (Tuple[int,int]): Second span

    Returns:
        bool: True if intersect, otherwise False.
    """
    return a[0]<=b[1] and b[0]<=a[1]

def is_span_nested(a: Tuple[int,int], b: Tuple[int,int])->bool:
    """
    Determine if two spans nested.
    (b[0]<=a[0] and a[1]<=b[1]) or (a[0]<=b[0] and b[1]<=a[1])

    Args:
        a (Tuple[int,int]): First span
        b (Tuple[int,int]): Second span

    Returns:
        bool: True if nested, otherwise False.
    """
    return (b[0]<=a[0] and a[1]<=b[1]) or (a[0]<=b[0] and b[1]<=a[1])

def decode_sequence_labels(sequence: List[str], offset=0, delimiter="-"):
    """
    decode sequence labels.

    Args:
        sequence (List[str]): Labels sequence.
        offset (int, optional): The offset of start position. Defaults to 0.
        delimiter (str, optional): The delimiter of labels sequence. Defaults to "-".

    Raises:
        ValueError: Invalid labels

    Returns:
        entities (Set[Tuple[int, int, str]]): The set of entities. The range is [l,r].
    """
    entities = set()
    start = -1
    entity = None
    for end in range(len(sequence)):
        if sequence[end].startswith("B") or sequence[end].startswith("S"):
            if start>=0 and entity is not None:
                entities.add((start-offset, end-offset-1, entity))
            start = end
            entity = delimiter.join(sequence[end].split(delimiter)[1:])
        elif sequence[end].startswith("M") or sequence[end].startswith("I"):
            continue
        elif sequence[end].startswith("E"):
            if start>=0 and entity is not None:
                entities.add((start-offset, end-offset, entity))
            start = -1
            entity = None
        elif sequence[end].startswith("O"):
            if start>=0 and entity is not None:
                entities.add((start-offset, end-offset-1, entity))
            start = -1
            entity = None
        else:
            raise ValueError(f"Invalid label {sequence[end]}")
    if start>=0 and entity is not None:
        entities.add((start-offset, len(sequence)-offset-1, entity))
    return entities


import random
import os
import sys
import importlib

def set_seed(seed: int = 42):
    """
    set seed

    Args:
        seed (int, optional): the seed of random. Defaults to 42.
    """
    random.seed(seed)
    os.environ['PYTHONHASHSEED']=str(seed)
    os.environ["CUBLAS_WORKSPACE_CONFIG"]=":16:8"
    try:
        import numpy as np
        np.random.seed(seed)
    except ImportError:
        pass
    try:
        import torch
        torch.manual_seed(seed)
        torch.cuda.manual_seed(seed)
        torch.cuda.manual_seed_all(seed)
        torch.backends.cudnn.deterministic = True
        torch.backends.cudnn.benchmark = False
        torch.backends.cudnn.enabled = False
        # scatter_add_cuda_kernel does not have a deterministic implementation
        # torch.use_deterministic_algorithms(True)
    except ImportError:
        pass
  
def is_in_notebook():
    """
    Check if in notebook environment

    Returns:
        bool: True if it is in a notebook, otherwise False.
    """
    try:
        # Test adapted from tqdm.autonotebook: https://github.com/tqdm/tqdm/blob/master/tqdm/autonotebook.py
        get_ipython = sys.modules["IPython"].get_ipython
        if "IPKernelApp" not in get_ipython().config:
            raise ImportError("console")
        if "VSCODE_PID" in os.environ:
            raise ImportError("vscode")
        if "DATABRICKS_RUNTIME_VERSION" in os.environ and os.environ["DATABRICKS_RUNTIME_VERSION"] < "11.0":
            # Databricks Runtime 11.0 and above uses IPython kernel by default so it should be compatible with Jupyter notebook
            # https://docs.microsoft.com/en-us/azure/databricks/notebooks/ipython-kernel
            raise ImportError("databricks")

        return importlib.util.find_spec("IPython") is not None
    except (AttributeError, ImportError, KeyError):
        return False  


import json
import os
from typing import Any, Dict, List, Tuple, Union
from tqdm import tqdm

def read_conll(fp: str, label_delimiter:str = "-", encoding:str='utf-8'):
    """
    Read conll-2003 BMES file
    
    Format:
    
    北 B-LOC
    京 I-LOC
    欢 O
    迎 O
    你 O

    Args:
        fp (str): The relative or absolute file path. 
        label_delimiter (str, optional): Default symbom of the delimiter. Defaults to "-".
        encoding (str, optional): The encoding of file. Defaults to "utf-8".

    Returns:
        Tuple[List,Set]: Return the sentence and the list of entity types within the boundary [l, r).
    """
    sentences = []
    entity_type_set = set()
    def init_sentence():
        return [],[]
    def init_entity():
        return {
            "start": -1,
            "end": -1,
            "entity": "",
            "text": []
        }
    sentence, entities = init_sentence()
    entity = init_entity()
    with open(fp, "r", encoding=encoding) as f:
        for line in f:
            line = line.strip().split()
            if len(line)==2 or (len(line)>0 and line[-1][0] in ("B", "M", "I", "E", "S")):
                if len(line)==2:
                    ch, label = line 
                else:
                    # case ' ' O
                    ch, label = ' ', line[-1]
                sentence.append(ch)
                if label.startswith("B") or label.startswith("S"):
                    if entity["end"]!=-1:
                        entities.append(entity)
                    entity = init_entity()
                    entity["start"] = len(sentence) - 1
                    entity["end"] = len(sentence)
                    entity["text"].append(ch)
                    entity["entity"] = label_delimiter.join(label.split(label_delimiter)[1:])
                    if label.startswith("S"):
                        entities.append(entity)
                        entity = init_entity()
                elif label.startswith("I") or label.startswith("M"):
                    if entity["start"]==-1:
                        entity["start"] = len(sentence)-1
                    entity["text"].append(ch)
                    entity["end"] = len(sentence)
                    if entity["entity"]=="":
                        entity["entity"] = label_delimiter.join(label.split(label_delimiter)[1:])
                elif label.startswith("E"):
                    if entity["start"]==-1:
                        entity["start"] = len(sentence)-1
                    entity["text"].append(ch)
                    entity["end"] = len(sentence)
                    if entity["entity"]=="":
                        entity["entity"] = label_delimiter.join(label.split(label_delimiter)[1:])
                    entities.append(entity)
                    entity = init_entity()
                elif label.startswith("O"):
                    if entity["end"]!=-1:
                        entities.append(entity)
                        entity = init_entity()
                else:
                    raise NotImplementedError()
            else:
                # next sentence
                if len(sentence)>0:
                    sentences.append({
                        "text": sentence,
                        "entities": entities
                    })
                    # fix: variable "entity" leak
                    for _entity in entities:
                        entity_type_set.add(_entity["entity"])
                entity = init_entity()
                sentence, entities = init_sentence()
        # last entity
        if entity["end"]!=-1:
            entities.append(entity)
        if len(sentence)>0:
            sentences.append({
                "text": sentence,
                "entities": entities
            })
            # fix: variable "entity" leak
            for _entity in entities:
                entity_type_set.add(_entity["entity"])
    return sentences, list(sorted(entity_type_set))

def read_weibo(fp:str, label_delimiter:str = "-", encoding:str='utf-8'):
    """
    Read original weibo dataset.
    
    Format:
    
    北0 B-LOC
    京1 I-LOC
    欢0 O
    迎0 O
    你0 O

    Args:
        fp (str): The relative or absolute file path. 
        label_delimiter (str, optional): Default symbom of the delimiter. Defaults to "-".
        encoding (str, optional): The encoding of file. Defaults to "utf-8".

    Returns:
        Tuple[List,Set]: Return the sentence and the list of entity types within the boundary [l, r).
    """
    sentences, entity_type_set = read_conll(fp, label_delimiter, encoding)
    for i in range(len(sentences)):
        item = {
            "text":list(ch[:-1] for ch in sentences[i]["text"]),
            "entities":[]
        }
        for entity in sentences[i]["entities"]:
            item["entities"].append(entity)
            item["entities"][-1].update({
                "text": list(ch[:-1] for ch in entity["text"]),
            })
        sentences[i] = item
    return sentences, entity_type_set


def save_jsonl(fp: str, data: List[Any], encoding='utf-8'):
    """
    Save as jsonl file.

    Args:
        fp (str): The filename
        data (List[Any]): The list of data
    """
    dirname = os.path.dirname(fp)
    os.makedirs(dirname, exist_ok=True)
    with open(fp,"w", encoding=encoding) as f:
        for item in data:
            f.write(json.dumps(item, ensure_ascii=False))
            f.write('\n')
            f.flush()
            
def save_labels(fp: str, data: List[str], encoding='utf-8'):
    """
    Save labels as plaintext.

    Args:
        fp (str): The filename
        data (List[str]): Labels
        encoding (str, optional): The encoding of file. Defaults to 'utf-8'.
    """ 
    dirname = os.path.dirname(fp)
    os.makedirs(dirname, exist_ok=True)
    with open(fp, "w", encoding=encoding) as f:
        for item in data:
            f.write(item)
            f.write('\n')
            f.flush()

